from __future__ import absolute_import

import logging
import os
import subprocess
import sys
import urllib

from conversion.boolean import convert_bool

from .config import API_URL, get_config
from .websocket import websocket


WEBSOCKET_URL = os.environ.get('WEBSOCKET_URL', 'wss://ws.zymbit.com/channel')
SSL_CHECK_HOSTNAME = convert_bool(os.environ.get('SSL_CHECK_HOSTNAME', 'true'))


def get_api_url(path):
    _config = get_config()
    api_url = _config.get('cloud', {}).get('api_url', API_URL)

    return '{}{}'.format(api_url, path)


def get_connection(url=None, credentials=None, check_hostname=None):
    headers = []

    logger = logging.getLogger(__name__)

    if check_hostname is None:
        check_hostname = SSL_CHECK_HOSTNAME

    sslopt = {"check_hostname": check_hostname}

    url = url or WEBSOCKET_URL
    request_url = url
    if credentials:
        encoded = urllib.urlencode(credentials)
        request_url = '{}?{}'.format(url, encoded)

    logger.debug('getting websocket connection at url={}'.format(url))

    ws = websocket.create_connection(request_url, header=headers, sslopt=sslopt)
    ws.settimeout(0)

    return ws


def get_system():
    # no support for non-linux systems
    if os.path.exists('/proc/cpuinfo'):
        content = open('/proc/cpuinfo', 'rb').read()
        if 'Arduino Yun' in content:
            return 'arduino'
        elif 'BCM270' in content:  # note this will match BCM2708 (rpi) and BCM2709 (rpi2)
            return 'raspberrypi'

    if sys.platform == 'darwin':
        return sys.platform


def run_command(command):
    return subprocess.Popen(command).wait()



import json
import os

API_URL = os.environ.get('API_URL', 'https://zymbit.com/api/v1')
CONFIG_PATH = os.environ.get('CONFIG_PATH', '/etc/zymbit.conf')
PROVISION_PATH = os.environ.get('PROVISION_PATH', '/provision')

TUNNEL_SSH_KEY_PATH = os.path.expanduser(os.environ.get('TUNNEL_SSH_KEY_PATH', '~/.ssh/id_tunnel'))

INIT_SCRIPT_PATH = '/etc/init.d/zymbit'
FILES_ROOT = None


def get_config():
    if os.path.exists(CONFIG_PATH):
        return json.load(open(CONFIG_PATH, 'rb'))

    return {}


def save_config(config):
    json.dump(config, open(CONFIG_PATH, 'wb'))



import datetime
import json
import logging

from zymbit.exceptions import ConnectionError

from zymbit.timeutil import LONG_TIME_AGO, now, timestamp

from zymbit.compat.console import Console


# interval to send a message that connection is down
CONNECTION_ERROR_INTERVAL = datetime.timedelta(seconds=30)


class ConsoleData(object):
    def __init__(self, read_callback=None):
        self.last_acquire = LONG_TIME_AGO

        self.console = Console()

        self._read_callback = read_callback or self.read_callback

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def read_callback(self, line):
        try:
            reading = json.loads(line)
        except ValueError, exc:
            self.logger.error('Unable to parse JSON: {!r}'.format(line))
        else:
            reading['timestamp'] = timestamp()
            return reading

    def get_readings(self):
        _now = now()
        readings = []

        try:
            payload = self.console.recv()
        except ConnectionError, exc:
            if _now - self.last_acquire >= CONNECTION_ERROR_INTERVAL:
                self.logger.exception(exc)
                self.logger.error('Unable to get data from arduino since {}'.format(
                    timestamp(self.last_acquire)))

            return

        self.last_acquire = _now

        self.logger.debug('payload={!r}'.format(payload))

        for line in payload.strip().splitlines():
            reading = self._read_callback(line)
            if reading is not None:
                readings.append(reading)

        return readings

    def loop(self):
        return self.get_readings()



import logging
import os

from subprocess import Popen


def create_ssh_key(path):
    """
    Create an ssh key.

    :param path: string - location for the ssh key
    :return: string - public ssh key
    """
    if os.path.exists(path):
        raise OSError('ssh key at path={} exists'.format(path))

    command = ('/usr/bin/ssh-keygen', '-t', 'rsa', '-b', '2048', '-f', path, '-N', '')
    logging.debug('command={}'.format(command))

    status = Popen(command).wait()
    assert status == 0

    return get_pubkey(path)


def get_pubkey(path):
    """
    Return the public key for the given SSH key

    :param path: - string path to the private SSH key
    :return: string - public ssh key
    """
    pub_key_path = '{}.pub'.format(path)
    with open(pub_key_path, 'rb') as fh:
        pub_key = fh.read()

    return pub_key.strip()



import collections
import json
import socket
import threading

AGGREGATES_LOCK = threading.Lock()


class MetricsAggregator(object):
    """
    Collects and groups
    """
    max_age = 1.0  # seconds
    value_names = {
        'disk': ('written', 'read'),
        'interface': ('tx', 'rx'),
        'load': ('1min', '5min', '15min'),
    }

    def __init__(self):
        # dict of string:metric_name, dict:data pairs
        # the data dict will contain a `time` key indicating the time the first entry came in
        self.aggregates = collections.OrderedDict()

    def aggregate(self, metric):
        ready = []

        _time = metric.time

        with AGGREGATES_LOCK:
            # collect metrics that are ready to be shipped
            keys = self.aggregates.keys()
            for key in keys:
                if self.is_ready(self.aggregates[key], _time):
                    ready.append(self.aggregates.pop(key))
                else:
                    break  # this is a sorted dict, so if the top isn't ready, later items won't be either

            basename = name = metric.type

            instance = metric.plugin_instance
            if instance:
                name = '{}.{}'.format(basename, instance)

            aggregate = self.aggregates.setdefault(name, {
                'name': name,
                'time': _time
            })

        # make sure the time in the aggregate is the first timestamp in this series
        aggregate['time'] = min(aggregate['time'], _time)

        metric_name = metric.type_instance or metric.plugin
        aggregate.update(self.get_values(metric_name, metric.values))

        return ready

    def get_values(self, metric_name, values):
        value_names = self.value_names.get(metric_name, [metric_name])
        return dict(zip(value_names, values))

    def is_ready(self, aggregate, time):
        return aggregate['time'] + self.max_age < time


class MetricsMessenger(object):
    """
    Ships metrics to the local messenger service
    """
    config = {}

    def __init__(self, aggregator=None):
        self._addr = None
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self.aggregator = aggregator or MetricsAggregator()

    @property
    def addr(self):
        if self._addr:
            return self._addr

        self._addr = (
            self.config.get('host', 'localhost'),
            self.config.get('port', 9628),
        )

        return self._addr

    def write(self, buf):
        for metric in self.aggregator.aggregate(buf):
            prefix = self.config.get('prefix')
            if prefix:
                metric['name'] = '{}.{}'.format(prefix, metric['name'])

            self.sock.sendto(json.dumps(metric), self.addr)


MESSENGER = MetricsMessenger()


def config(config_item):
    """
    child: host, values: ('localhost',)
    child: port, values: ('8001',)
    """
    MESSENGER.config.update([(x.key, x.values[0]) for x in config_item.children])


def write(metric, data=None):
    MESSENGER.write(metric)



from zymbit.util import *


def get_distro():
    distro = None

    issue_path = '/etc/issue'
    if os.path.exists(issue_path):
        with open(issue_path, 'rb') as fh:
            content = fh.read().lower()
            for _distro in ('raspbian', 'arch'):
                if _distro in content:
                    distro = _distro
                    break

    return distro



import logging
import socket

from select import select
from zymbit.connector import get_connector
from zymbit.envelope import parse_buf

BUFSIZE = 4096


class MessengerServer(object):
    def __init__(self, host, port, message_handler=None):
        self.addr = (host, port)

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.sock.setblocking(0)

        self.sock.bind(self.addr)
        self.sock.listen(128)  # max 128 clients

        self.logger.info("Listening on TCP addr={}".format(self.addr))

        self.udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.udp_sock.bind(self.addr)

        self.logger.info("Listening on UDP addr={}".format(self.addr))

        self.connections = {}

        self.message_handler = message_handler

        self._run = True

    @property
    def logger(self):
        logger_name = '{}.{}'.format(__name__, self.__class__.__name__)
        return logging.getLogger(logger_name)

    def connect(self, info):
        conn, addr = info
        self.logger.info('%s, %s connected' % (conn, addr))

        self.connections[conn] = addr

    def disconnect(self, connection):
        addr = self.connections.pop(connection)
        self.logger.info('%s, %s disconnected' % (connection, addr))

    def fileno(self):
        return self.sock.fileno()

    def handle_message(self, client, buf):
        if self.message_handler:
            self.message_handler.handle_message(client, self.connections[client], buf)
        else:
            message = 'client={}, buf={}'.format(self.connections[client], buf)
            self.logger.info(message)

            client.send(message)

    def loop(self, select_timeout=1.0):
        # check UDP
        buf, client = self.udp_sock.recvfrom(1024)
        if buf:
            self.handle_message(client, buf)

        try:
            self.connect(self.sock.accept())
        except socket.error, exc:
            # (11, 'Resource temporarily unavailable')
            # [Errno 35] Resource temporarily unavailable
            if exc[0] not in (11, 35):
                raise

        ready, _, _ = select(self.connections, [], [], select_timeout)

        for client in ready:
            try:
                buf = client.recv(BUFSIZE)
                if buf == '':
                    self.disconnect(client)
                    continue
            except socket.error, exc:
                if exc[0] != 54: # [Errno 54] Connection reset by peer
                    raise

                self.disconnect(client)
                continue

            self.handle_message(client, buf)

        if self.message_handler:
            self.message_handler.loop()

    def quit(self):
        self.sock.close()

        # prevent getting exception where dictionary changes while looping
        connections = self.connections.keys()
        for connection in connections:
            self.disconnect(connection)

    def run(self):
        while self._run:
            self.loop()


class ConsoleMessengerServer(MessengerServer):
    port = 9628
    host = 'localhost'

    def __init__(self, host=None, port=None, message_handler=None):
        host = host or self.host
        port = port or self.port

        super(ConsoleMessengerServer, self).__init__(host, port, message_handler=message_handler)

    @property
    def connector(self):
        return get_connector()

    def handle_message(self, client, buf):
        self.connector.send(parse_buf(buf))

    def loop(self, select_timeout=0.1):
        return super(ConsoleMessengerServer, self).loop(select_timeout=select_timeout)


if __name__ == '__main__':
    import sys

    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

    server = MessengerServer('localhost', 9628)
    try:
        server.run()
    except KeyboardInterrupt:
        server.quit()



import logging
import select
import socket

from zymbit.exceptions import ConnectionError

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 6571


class Console(object):
    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT):
        self.host = host
        self.port = port

        self._socket = None

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def close(self):
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()

        self._socket = None

    def fileno(self):
        return self.socket.fileno()

    def handle_socket_exception(self, socket_exc):
        """
        Handle the socket exception
        :param socket_exc: Exception
        """

    @property
    def socket(self):
        if self._socket is not None:
            return self._socket

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.settimeout(1.0)

        try:
            self._socket.connect((self.host, self.port))
        except socket.error, exc:
            self.handle_socket_exception(exc)

            raise ConnectionError(
                'Unable to connect to host={}, port={}'.format(self.host, self.port))

        return self._socket

    def recv(self, timeout=2.0):
        socket_list = [self.socket]

        buf = ''
        while True:
            # Get the list sockets which are readable
            _read, _, _ = select.select(socket_list, [], [], timeout)
            if _read:
                # after getting the first byte, don't wait very long for more
                # this is so the arduino is given time to process the given
                # command, but once we start getting a response, timeout is quick
                timeout = 0.1
            else:
                break

            for _socket in _read:
                if _socket == self.socket:
                    data = _socket.recv(4096)
                    if not data:
                        self._socket = None

                        raise ConnectionError('lost connection')
                    else:
                        buf += data
                else:
                    raise Exception(
                        'Unknown _socket={}'.format(_socket))

        return buf

    def send(self, buf):
        try:
            self.socket.send(buf)
        except socket.error, exc:
            if exc[0] != 32:
                raise

            self._socket = None
            self.send(buf)



import logging

import zymbit.config

from zymbit.commands import Command
from zymbit.commands.util import filter_parser_args
from zymbit.tunnel import Tunnel


class TunnelCommand(Command):
    command_name = 'tunnel'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def run(self):
        parser_args = filter_parser_args(self.args)

        self.logger.info('parser_args={}'.format(parser_args))

        tunnel = Tunnel(**parser_args)

        try:
            while True:
                tunnel.check()
        except KeyboardInterrupt:
            tunnel.close()



class CommandError(Exception):
    pass


class ConnectionError(Exception):
    pass


class NotConnected(Exception):
    pass


class ProvisionError(Exception):
    pass


class TasksFailed(Exception):
    pass


class TunnelError(Exception):
    pass



import os
import pkgutil

from importlib import import_module

COMMANDS_DIR, UTIL_NAME = os.path.split(__file__)


def filter_parser_args(args, filter_none=True):
    parser_args = dict([(k, getattr(args, k)) for k in dir(args) if not k.startswith('_')])
    parser_args.pop('action')
    parser_args.pop('config', None)
    parser_args.pop('logfile')
    parser_args.pop('loglevel')

    if filter_none:
        keys = parser_args.keys()
        for key in keys:
            if parser_args[key] is None:
                parser_args.pop(key)

    return parser_args


def get_classes(package, subclass):
    """
    Generator that finds all classes of the given subclass

    :param package: an imported python package; e.g. import zymbit.commands
    :param subclass: subclass to search for in modules
    :return: generator
    """
    for module_path in get_modules(package):
        module = import_module(module_path)

        for name in dir(module):
            item = getattr(module, name)
            if item == subclass:
                continue

            try:
                if issubclass(item, subclass):
                    yield item
            except TypeError:
                pass


def get_modules(package):
    """
    find all 
    :param package: an imported python package; e.g. import zymbit.commands
    :return: list[string] - module paths
    """
    package_name = package.__name__
    path = package.__path__[0]
    modules = ['{}.{}'.format(package_name, name) for _, name, _ in pkgutil.iter_modules([path])]

    return modules



from __future__ import absolute_import

import os

from zymbit.util import run_command

from . import config


def create_ssh_key():
    command = ('/usr/bin/dropbearkey', '-t', 'rsa', '-s', '2048', '-f', config.TUNNEL_SSH_KEY_PATH)
    return run_command(command)


def enable_init_script():
    os.chmod(config.INIT_SCRIPT_PATH, 0755)

    command = (config.INIT_SCRIPT_PATH, 'enable')
    return run_command(command)


def start_service():
    command = (config.INIT_SCRIPT_PATH, 'start')
    return run_command(command)



import logging
import os
import shutil

from zymbit.compat import config, provision, util

SCRIPT_DIR = os.path.dirname(__file__)


class Provision(object):
    def __init__(self, token, api_url=None, websocket_url=None):
        self.token = token
        self.api_url = api_url
        self.websocket_url = websocket_url
        self.ssh_pubkey = None

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def copy_files(self):
        distro = util.get_distro()
        system = util.get_system()

        files_root = None
        _path = SCRIPT_DIR
        while _path != '/':
            system_files_path = os.path.join(_path, 'files', system)
            if os.path.exists(system_files_path):
                files_root = os.path.dirname(system_files_path)
                break

            _path = os.path.dirname(_path)

        if not files_root:
            self.logger.warning('files_root not found, skipping copy_files')
            return

        for tree in (system, '{}-{}'.format(system, distro)):
            tree_root = os.path.join(files_root, tree)

            if not os.path.exists(tree_root):
                self.logger.warning('tree_root={} does not exist'.format(tree_root))
                continue

            self.logger.info('copying files from tree_root={} to /'.format(tree_root))

            try:
                _makedirs = shutil.os.makedirs
                shutil.os.makedirs = lambda *args, **kwargs: None
                shutil.copytree(tree_root, '/')
            finally:
                shutil.os.makedirs = _makedirs

    def run(self):
        self.write_config()
        self.copy_files()

        system = util.get_system()

        try:
            status = provision.enable_init_script()
        except AttributeError:
            self.logger.warning('enable_init_script not found for system={}'.format(system))
        else:
            if status != 0:
                self.logger.warning('enable_init_script exited with status={}'.format(status))

        try:
            status = provision.start_service()
        except AttributeError:
            self.logger.warning('start_service not found for system={}'.format(system))
        else:
            if status != 0:
                self.logger.warning('start_service exited with status={}'.format(status))

        try:
            status = provision.start_metrics()
        except AttributeError:
            self.logger.warning('start_metrics not found for system={}'.format(system))
        else:
            if status != 0:
                self.logger.warning('start_metrics exited with status={}'.format(status))

    def write_config(self):
        _config = config.get_config()

        if self.websocket_url:
            _config.setdefault('cloud', {})['websocket_url'] = self.websocket_url

        if self.api_url:
            _config.setdefault('cloud', {})['api_url'] = self.api_url

        _config.setdefault('auth', {})['provisioning_token'] = self.token

        config.save_config(_config)



from zymbit.commands import Command
from zymbit.provision import Provision


class ProvisionCommand(Command):
    command_name = 'provision'

    @classmethod
    def setup_parser(cls, parser):
        command_parser = super(ProvisionCommand, cls).setup_parser(parser)

        command_parser.add_argument('--api-url')
        command_parser.add_argument('--websocket-url')
        command_parser.add_argument('token')

        return command_parser

    def run(self):
        token = self.args.token
        api_url = self.args.api_url
        websocket_url = self.args.websocket_url

        provision = Provision(token, api_url=api_url, websocket_url=websocket_url)
        provision.run()



import argparse
import logging

import zymbit  # trigger setting up import hooks

import zymbit.config

from zymbit.commands.util import get_classes
from zymbit.logconfig import config_logging


class Command(object):
    """
    Command subclass for all commands

    The get_classes() util will filter classes based on a subclass.
    This class should be used on all commands in order to include
    them.
    """
    command_name = 'subcommand'

    def __init__(self, parser, args):
        self.parser = parser
        self.args = args

    @classmethod
    def process_command(cls, package):
        parser = argparse.ArgumentParser()
        parser.add_argument('--logfile')  # default None
        parser.add_argument('--loglevel', '-l', default='info')
        parser.add_argument('--config', '-c')

        subparser = parser.add_subparsers(dest='action')

        commands = dict([(x.command_name, x) for x in get_classes(package, cls)])
        for command in commands.values():
            command.setup_parser(subparser)

        args = parser.parse_args()
        if args.config:
            zymbit.config.CONFIG_PATH = args.config

        # setup logging to info
        loglevel = getattr(logging, args.loglevel.upper())

        config_logging(args.logfile, loglevel)

        logger = logging.getLogger(__name__)

        logger.debug('action={}'.format(args.action))
        try:
            action_cls = commands[args.action]
        except KeyError:
            return 'unknown action={}'.format(args.action)

        return action_cls(parser, args).run()

    @classmethod
    def setup_parser(cls, parser):
        command_parser = parser.add_parser(cls.command_name)

        return command_parser

    def run(self):
        print 'run stub for command_name={}'.format(self.command_name)



import logging
import time

from zymbit.linux.console import Console

from subprocess import PIPE, STDOUT, Popen

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 6571


class ArduinoConsole(Console):
    @property
    def logger(self):
        return logging.getLogger(__name__)

    def _restart_bridge(self):
        """
        Checks to see if the bridge process is running
        :return: boolean - whether the bridge was restarted
        """
        running = self._is_bridge_running()
        if not running:
            self.logger.warning('bridge not running, resetting mcu')
            status = Popen('/usr/bin/reset-mcu').wait()
            self.logger.info('reset-mcu status={}'.format(status))

            for i in range(10):  # wait for bridge to show up
                running = self._is_bridge_running()
                if running:
                    break

                time.sleep(1.0)
            else:
                self.logger.warning('bridge still not found in process listing')

            if running:  # let the bridge settle for a couple seconds
                time.sleep(2.0)

            return running

        return False

    def handle_socket_exception(self, socket_exc):
        """
        Handle the socket exception
        :param socket_exc: Exception
        """
        # try to restart the bridge on arduinos and if successful, try reopening the socket
        if self._restart_bridge():
            return self.socket

    def _is_bridge_running(self):
        """
        Check to see if bridge is running
        :return: Boolean
        """
        command = 'ps'

        proc = Popen(command, stdout=PIPE, stderr=STDOUT)
        stdout, _ = proc.communicate('')

        running = 'python -u bridge.py' in stdout

        return running



import logging.config


def config_logging(logfile=None, loglevel=None):
    logging.config.dictConfig(get_logging_config(logfile, loglevel))


def get_logging_config(logfile=None, loglevel=None):
    logfile = logfile or '/var/log/zymbit.log'
    loglevel = loglevel or 'INFO'

    return {
        'version': 1,
        'disable_existing_loggers': True,

        'formatters': {
            'default': {
                'format': '%(levelname)s %(asctime)s %(name)s %(message)s',
                'datefmt': '%Y-%m-%d %H:%M:%S',
            },
        },

        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': 'DEBUG',
                'stream': 'ext://sys.stdout',
                'formatter': 'default',
            },
            'file': {
                'class': 'logging.handlers.RotatingFileHandler',
                'level': 'INFO',
                'formatter': 'default',
                'filename': logfile,
                'maxBytes': 1024 * 100,
                'backupCount': 2,
            },
            'null': {
                'class': 'logging.NullHandler',
                'level': 'DEBUG',
            },
        },

        'loggers': {
            '': {
                'handlers': ['console'],
                'propagate': False,
                'level': loglevel,
            },
            'filer': {
                'handlers': ['console', 'file'],
                'propagate': False,
                'level': loglevel,
            },
            'requests': {
                'handlers': ['null'],
                'propagate': False,
                'level': 'ERROR',
            },
        },
    }



from __future__ import absolute_import

import logging
import socket
import time

from zymbit.exceptions import NotConnected
from zymbit.websocket import WebSocketConnectionClosedException

from .client import Client


class Stream(Client):
    def __init__(self, subscriptions):
        self.subscriptions = subscriptions

        super(Stream, self).__init__(connected_callback=self.online)

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def handle(self, envelope):
        pass

    def loop(self):
        pass

    def online(self):
        """
        Called when websocket connection is initialized
        """
        self.logger.debug('connected, subscribing ...')
        for routing_key in self.subscriptions:
            params = {'routing_key': routing_key}
            self.logger.debug('subscribe params={}'.format(params))
            self.send('subscribe', params)

    def run(self):
        while True:
            try:
                envelope = self.recv()
            except socket.error as exc:
                if exc[0] == 11:
                    self.loop()
                    time.sleep(1.0)
                    continue

                self.logger.exception(exc)
                self.logger.error('socket closed, resetting connection')

                self._ws = None
            except NotConnected:
                time.sleep(1.0)
            except WebSocketConnectionClosedException:
                self._ws = None
            else:
                if envelope is None:
                    time.sleep(1.0)
                    continue

                self.handle(envelope)



import datetime
import json
import logging
import sys
import uuid

from zymbit.client import Client


def timestamp():
    return datetime.datetime.utcnow().isoformat('T')


def main(client_id, device_id, value):
    client = Client()

    envelope = json.dumps({
        'message_id': str(uuid.uuid4()),
        'client_id': client_id,
        'timestamp': timestamp(),
        'action': 'store',
        'message': [
            {'deviceId': device_id, 'timestamp': timestamp(), 'value': value},
        ],
    })

    client.ws.send(envelope)
    client.ws.close()

if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

    client_id = sys.argv[1]
    device_id = str(sys.argv[2])
    value = float(sys.argv[3])

    main(client_id, device_id, value)



from __future__ import absolute_import

import importlib
import logging
import os
import socket

from conversion import convert_bool

import requests

from requests.exceptions import HTTPError

from zymbit import config, statemachine, util
from zymbit.client import Client
from zymbit.clientid import get_client_id
from zymbit.compat import config
from zymbit.exceptions import NotConnected, TasksFailed
from zymbit.linux.util import create_ssh_key, get_pubkey

CONNECTION_REFUSED = 61

PROVISION_PATH = os.environ.get('PROVISION_PATH', '/provision')


def get_connector():
    return Connector.instance()


class Connector(statemachine.StateMachine):
    _instance = None

    def __init__(self):
        super(Connector, self).__init__()

        self.client = None
        self.system = util.get_system()

        # tasks that should be run on loop
        self.tasks = {}
        self.tasks_config = {}

    @classmethod
    def instance(cls):
        if cls._instance:
            return cls._instance

        cls._instance = cls()

        return cls._instance

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def connect(self):
        if not self.client:
            self.client = Client()

        # don't reconnect an already-connected client
        connected = self.client.is_connected
        if connected:
            return True

        try:
            self.client.connect()
        except NotConnected:
            pass
        else:
            connected = True

        if connected:
            self.logger.info('connected')
            return True

        self.load_tasks()

        # process tasks if not connected
        self.process_tasks()

    def connected(self):
        try:
            line = self.client.recv()
        except socket.error, exc:
            if exc[0] not in (110, 111):
                raise

            self.client.close()
            raise NotConnected
        else:
            if line:
                print 'line={}'.format(line)

        self.process_tasks()

    def create_ssh_key(self):
        _config = config.get_config()
        tunnel_config = _config.setdefault('tunnel', {})
        identity_file = tunnel_config.get('identity_file', config.TUNNEL_SSH_KEY_PATH)
        if not os.path.exists(identity_file):
            create_ssh_key(identity_file)
            self.logger.info('created ssh key at identity_file={}'.format(identity_file))

        # check to see if config should be written
        if not tunnel_config.get('host'):
            _config['tunnel'].update({
                'identity_file': identity_file,
                'user': 'tunnel',
                'host': 'tunnel.zymbit.com',
            })
            config.save_config(_config)

        return True

    def get_auth(self):
        _config = config.get_config()
        client_id = get_client_id()

        return _config.get('auth', {}).get(client_id)

    def get_provisioning_token(self):
        _config = config.get_config()

        return _config.get('auth', {}).get('provisioning_token')

    def load_task(self, name, task_config):
        module_path, class_name = task_config['class_path'].rsplit('.', 1)

        task_module = importlib.import_module(module_path)
        task_class = getattr(task_module, class_name)

        task = task_class(**task_config.get('params', {}))

        # print 'task name={}, task_config={}, task_class={}, task={}'.format(name, task_config, task_class, task)
        self.tasks[name] = task

    def load_tasks(self):
        self.logger.debug('load_tasks')

        _config = config.get_config()
        self.tasks_config = _config.get('tasks', {})
        for name, task_config in self.tasks_config.items():
            # if a task is already loaded, skip
            if name in self.tasks:
                continue

            try:
                self.load_task(name, task_config)
            except Exception, exc:
                self.logger.exception(exc)
                self.logger.error('task name={} raised exception when trying to load it'.format(name))

        return True

    def process_tasks(self):
        reload_tasks = False
        task_names = self.tasks.keys()
        for task_name in task_names:
            task = self.tasks[task_name]

            try:
                # self.logger.debug('task task_name={}, task={}'.format(task_name, task))
                task.loop()
            except Exception, exc:
                self.logger.exception(exc)
                self.logger.error('task_name={} raised exception above'.format(task_name))

                self.quit_task(task_name, task)
                reload_tasks = True

        if reload_tasks:
            raise TasksFailed

    def quit(self):
        for name, task in self.tasks.items():
            self.quit_task(name, task)

    def quit_task(self, name, task):
        try:
            task.quit()
        except Exception, exc2:
            self.logger.exception(exc2)
            self.logger.error('task name={} raised exception above trying to quit'.format(name))

        self.tasks.pop(name, None)  # don't barf if the task is not in the tasks list

    def reload_tasks(self):
        return self.load_tasks()

    def register(self):
        client_id = get_client_id()
        auth = self.get_auth()
        if auth:
            return True

        provisioning_token = self.get_provisioning_token()
        if not provisioning_token:
            self.logger.debug('waiting for provisioning token')
            return

        self.logger.debug('provisioning, client_id={}'.format(client_id))

        client_id = get_client_id()
        provision_url = util.get_api_url(config.PROVISION_PATH)

        ssh_pubkey = get_pubkey(config.TUNNEL_SSH_KEY_PATH)

        data = {
            'token': provisioning_token,
            'client_id': client_id,
            'ssh_pubkey': ssh_pubkey,
        }

        verify = convert_bool(os.environ.get('SSL_CHECK_HOSTNAME', 'true'))

        response = requests.post(provision_url, data=data, verify=verify)

        try:
            response.raise_for_status()
        except HTTPError, exc:
            self.logger.exception(exc)
            self.logger.error(response.content)

            raise

        response_data = response.json()

        self._write_config(response_data, config.TUNNEL_SSH_KEY_PATH)

    def run(self):
        while self._run:
            try:
                super(Connector, self).run()
            except (NotConnected, TasksFailed):
                pass

    def send(self, envelope):
        self.client.send_raw(envelope)

    def _write_config(self, response_data, identity_file):
        _config = config.get_config()
        _config.setdefault('auth', {}).update({
            response_data['client_id']: {
                'uuid': response_data['uuid'],
                'auth_token': response_data['auth_token'],
            },
        })

        # save the SSH tunnel identity file
        _config.setdefault('tunnel', {}).update({
            'identity_file': identity_file,
            'ssh_port': response_data['ssh_port'],
        })

        config.save_config(_config)

    # now that the methods are defined, use the function definitions themselves
    # to setup the state machine transitions
    transitions = {
        connect: {
            True: connected,
        },
        connected: {
            NotConnected: connect,
            TasksFailed: reload_tasks,
        },
        create_ssh_key: {
            True: load_tasks,
        },
        load_tasks: {
            True: register,
        },
        reload_tasks: {
            True: connect,
        },
        register: {
            True: connect,
        },
        statemachine.StateMachine.start: {
            True: create_ssh_key,
        },
    }



#!/usr/bin/env python
import logging
import json
import sys

import dateutil.parser

from zymbit.stream import Stream


class EchoStream(Stream):
    def handle(self, envelope):
        data = json.loads(envelope)

        timestamp = dateutil.parser.parser(data['timestamp'])
        value = data['params'].get('value')

        print('envelope={}, timestamp={}, value={}'.format(
            envelope, timestamp, value))


if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

    subscriptions = sys.argv[1:]

    if not subscriptions:
        sys.exit('No subscriptions made, exiting')

    EchoStream(subscriptions).run()



import logging

from zymbit.commands import Command
from zymbit.connector import get_connector


class ConnectorCommand(Command):
    command_name = 'connector'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def run(self):
        connector = get_connector()

        try:
            while True:
                connector.run()
        except KeyboardInterrupt:
            connector.quit()



import datetime
import logging
import os
import select
import shutil
import time

from subprocess import PIPE, STDOUT, Popen

from zymbit.compat import config
from zymbit.exceptions import TunnelError
from zymbit.timeutil import now

SSH = os.environ.get('SSH', '/usr/bin/ssh')
MAX_BACKOFF = 300  # 5 minutes

# how long a connection must be in place in order to consider it okay to clear backoff
CONNECTION_OK_DELTA = datetime.timedelta(seconds=30)


class Tunnel(object):
    defaults = {
        'user': None,
        'host': None,
        'identity_file': None,
        'ssh_port': None,
        'port': 22,
    }

    def __init__(self, **kwargs):
        _config = self.defaults.copy()
        _config.update(config.get_config().get('tunnel', {}))
        _config.update(dict([item for item in kwargs.items() if item[1]]))

        unconfigured = [key for key, val in _config.items() if val is None]
        if unconfigured:
            raise TunnelError('{} cannot be unset'.format(unconfigured))

        self.user = _config['user']
        self.host = _config['host']
        self.port = str(_config['port'])
        self.identity_file = _config['identity_file']
        self.ssh_port = str(_config['ssh_port'])

        self.connection = None
        self.connected_at = None

        self.backoff = 0
        self.backoff_until = None

        self.connect()

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def close(self):
        if self.connection is None:
            return

        try:
            self.connection.kill()
        except OSError, exc:
            if exc[0] != 3:
                raise

        self.connection.wait()

        self.connection = None

    def check(self):
        if not self.connection:
            self.connect()
            if self.connection is None:
                self.logger.debug('no connection, sleeping 1.0s')
                time.sleep(1.0)
                return

        # this is the main throttling sleep here, keep it at around 1 second
        stdout = self.connection.stdout
        r, _, _ = select.select([stdout], [], [], 1.0)
        for stream in r:
            line = stream.readline().strip()
            if line:
                self.logger.info('line={}'.format(line))

        status = self.connection.poll()
        if status is None:
            # connection is established, see if backoff needs to be reset
            if self.backoff and now() > self.connected_at + CONNECTION_OK_DELTA:
                self.logger.debug('connection ok, resetting backoff')
                self.backoff = 0
                self.backoff_until = None

            return

        # a connection had been established, but has ended, need to reconnect
        self.logger.warning('closing connection after status={}'.format(status))
        self.close()

        # set the backoff time
        self.backoff += 1
        seconds = min(2**self.backoff, MAX_BACKOFF)
        self.backoff_until = now() + datetime.timedelta(seconds=seconds)

        self.logger.debug('backoff={}, seconds={}, backoff_until={}'.format(
            self.backoff, seconds, self.backoff_until))

    def check_host_key(self):
        zymbit_known_hosts = config.ZYMBIT_KNOWN_HOSTS
        known_hosts = config.KNOWN_HOSTS
        if not os.path.exists(known_hosts):
            shutil.copyfile(zymbit_known_hosts, known_hosts)
            return True

        with open(zymbit_known_hosts, 'rb') as zbkh:
            while True:
                pubkey = zbkh.readline()
                if pubkey == '':  # end of file
                    break

                pubkey = pubkey.strip()
                if pubkey == '':  # skip empty lines
                    continue

                known_hosts_temp = '{}.tmp'.format(known_hosts)

                try:
                    line_found = False
                    with open(known_hosts_temp, 'wb') as new_hosts:
                        with open(known_hosts, 'rb') as old_hosts:
                            line = old_hosts.readline().strip()

                            if line == pubkey:
                                line_found = True

                            new_hosts.write('{}\n'.format(line))

                        if not line_found:
                            new_hosts.write('{}\n'.format(pubkey))

                    # overwrite the temporary file
                    shutil.move(known_hosts_temp, known_hosts)
                finally:
                    if os.path.exists(known_hosts_temp):
                        os.remove(known_hosts_temp)

    def connect(self):
        if self.backoff_until and self.backoff_until > now():
            self.logger.debug('not connecting, backoff_until={}'.format(self.backoff_until))
            return

        self.check_host_key()

        self.connection = Popen((
            SSH, '-N', '-I', '60', '-R', self.tunnel_conf,
            '-i', self.identity_file, '-p', self.port, self.remote_server
        ), stdout=PIPE, stderr=STDOUT)

        self.connected_at = now()

        return self.connection

    def loop(self):
        return self.check()

    def reconnect(self):
        self.close()
        self.connect()

    @property
    def remote_server(self):
        return '{}@{}'.format(self.user, self.host)

    @property
    def tunnel_conf(self):
        return '{}:localhost:22'.format(self.ssh_port)



import logging
import time

from zymbit.timeutil import now, get_sleep_time


class StateMachine(object):
    transitions = {}

    def __init__(self):
        self._run = True
        self._state = self.start

        self._setup_transitions()
        self.logger.debug('transitions={}'.format(self.transitions))

    def _setup_transitions(self):
        # convert the transition functions into bound methods
        _transitions = {}
        for k, v in self.transitions.items():
            bound_method = getattr(self, k.func_name)
            t_transitions = dict([(kk, getattr(self, vv.func_name)) for kk, vv in v.items()])

            _transitions[bound_method] = t_transitions

        self.transitions = _transitions

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def get_new_state(self, transition_key):
        transitions = self.transitions.get(self._state, {})

        return transitions.get(transition_key)

    def loop(self):
        exc = None
        result = None

        try:
            result = self._state()
        except Exception, exc:  # global exception catcher here to use for state transitions
            self._state = self.get_new_state(exc) or self._state

            raise

        transitions = self.transitions.get(self._state, {})
        if result:
            self._state = transitions.get(result) or self._state

        return result

    def quit(self):
        self._run = False

    def run(self):
        while self._run:
            start = now()

            result = self.loop()
            if result is None:
                time.sleep(get_sleep_time(1.0, start))

    def start(self):
        return True



import sys

from zymbit.util import get_system


class CompatLoader(object):
    def load_module(self, fullname):
        system = get_system()
        system_name = fullname.replace('.compat.', '.{}.'.format(system))

        module = __import__(system_name)
        for _module_name in system_name.split(".")[1:]:
            module = getattr(module, _module_name)

        return module


class CompatFinder(object):
    def __init__(self, path_entry):
        if not path_entry.endswith('zymbit/compat'):
            raise ImportError()

    def find_module(self, fullname, path=None):
        return CompatLoader()
sys.path_hooks.append(CompatFinder)
